from typing import Optional

import numpy as np
from numba import njit
from rich.console import Console

VALUE = 0
SUIT = 1
TOTAL_CARDS = 2

PRETTY_VALUES = {
    0: "A",
    1: "2",
    2: "3",
    3: "4",
    4: "5",
    5: "6",
    6: "7",
    7: "8",
    8: "9",
    9: "10",
    10: "J",
    11: "Q",
    12: "K",
}

PRETTY_SUITS = {0: "♣", 1: "♠", 2: "♥", 3: "♦"}

SUIT_STYLES = {
    0: "bold black on bright_white",
    1: "bold black on bright_white",
    2: "bold dark_red on bright_white",
    3: "bold dark_red on bright_white",
}


NEW_DECK = np.array([(face, suit) for face in range(13) for suit in range(4)], np.uint8)

console = Console()


def print_current_state(
    game,
    game_index,
    card_being_checked: Optional[int] = None,
    last_stacked: Optional[int] = None,
):
    total = game[0][TOTAL_CARDS]
    index = 0
    board = []

    while total > 0:
        value = PRETTY_VALUES[game[index][VALUE]]
        suit = PRETTY_SUITS[game[index][SUIT]]
        style = SUIT_STYLES[game[index][SUIT]]
        total = game[index][TOTAL_CARDS]
        selection = ""
        stacked = ""
        if card_being_checked == index:
            selection = "[default on blue] [/default on blue]"
        if last_stacked == index:
            stacked = "[default on red] [/default on red]"
        board.append(
            f"{stacked}{selection}"
            f"[{style}] {value}{suit} [/{style}]"
            f"{selection}{stacked}"
            f"([green]{total}[/green])"
        )
        index += 1
        total = game[index][TOTAL_CARDS]

    console.print(" ".join(board))
    console.print(f"Number of active cards: {game_index + 1}")


@njit
def compare(array: np.array, index1: int, index2: int) -> bool:
    return (
        array[index1][VALUE] == array[index2][VALUE]
        or array[index1][SUIT] == array[index2][SUIT]
    )


@njit
def move_card(array: np.array, old_index: int, new_index: int):
    array[old_index][VALUE] = array[new_index][VALUE]
    array[old_index][SUIT] = array[new_index][SUIT]
    array[old_index][TOTAL_CARDS] += array[new_index][TOTAL_CARDS]

    array[new_index][VALUE] = 0
    array[new_index][SUIT] = 0
    array[new_index][TOTAL_CARDS] = 0


@njit
def shift(array, index):
    card_total = array[index + 1][TOTAL_CARDS]

    while card_total > 0:
        move_card(array, index, index + 1)
        index += 1
        card_total = array[index + 1][TOTAL_CARDS]


@njit
def get_shuffled_deck() -> np.array:
    deck = np.copy(NEW_DECK)
    np.random.shuffle(deck)
    return deck





from typing import Optional

import numpy as np
from numba import njit

from cg_simulator.core import compare, move_card, shift, VALUE, SUIT, TOTAL_CARDS, get_shuffled_deck, print_current_state, console, SUIT_STYLES


@njit
def _compare_replace(array, index) -> tuple[int, bool]:
    if index >= 3:
        if compare(array, index, index - 3):
            move_card(array, index - 3, index)
            shift(array, index)
            return index - 3, True
    if index > 0:
        if compare(array, index, index - 1):
            move_card(array, index - 1, index)
            shift(array, index)
            return index - 1, True

    return index, False


@njit
def _move_from_old_to_new(
    old_array: np.array, new_array: np.array, old_index: int, new_index: int
):
    new_array[new_index][VALUE] = old_array[old_index][VALUE]
    new_array[new_index][SUIT] = old_array[old_index][SUIT]
    new_array[new_index][TOTAL_CARDS] = 1


def _check_validity(array) -> bool:
    index = 51
    while index >= 0:
        if array[index][TOTAL_CARDS] == 0:
            index -= 1
            continue

        if index > 0 and compare(array, index, index - 1):
            return False
        if index >= 3 and compare(array, index, index - 3):
            return False

        return True


@njit
def _combine(array, index, game_index):
    replace_index, replaced = _compare_replace(array, index)

    if not replaced:
        return game_index

    game_index = game_index - 1
    index = game_index
    target_index = replace_index

    while target_index <= index:
        replace_index, replaced = _compare_replace(array, index)
        if replaced:
            target_index = replace_index
            game_index -= 1
            index = game_index
        else:
            index -= 1

    return game_index


@njit
def simulate() -> int:
    game = np.zeros((52, 3), np.uint8)
    deck = get_shuffled_deck()
    _move_from_old_to_new(deck, game, 0, 0)
    game_index = 1
    for i in range(1, 52):
        _move_from_old_to_new(deck, game, i, game_index)
        game_index = _combine(game, game_index, game_index) + 1

    return game[0][TOTAL_CARDS]


@njit
def multiple_simulations(n: int):
    results = np.zeros(n)
    for i in range(n):
        results[i] = simulate()

    return results


def combine_recursive(
    game: np.array,
    game_index: int,
    index: int,
    target_index: int,
    interactive: bool = False,
    last_stacked: Optional[int] = None,
) -> int:
    if interactive:
        print_current_state(
            game,
            game_index,
            card_being_checked=index,
            last_stacked=last_stacked,
        )
        input("Press enter to continue")
        print()
    replace_index, replaced = _compare_replace(game, index)

    if replaced:
        game_index -= 1
        return combine_recursive(
            game, game_index, game_index, replace_index, interactive, replace_index
        )
    elif replace_index > target_index:
        return combine_recursive(
            game, game_index, index - 1, target_index, interactive, last_stacked
        )

    return last_stacked, game_index


def simulate_recursive(
    print_game: bool = False,
    interactive: bool = False,
    save_csv: Optional[str] = False,
    deck_csv: Optional[str] = None,
):
    game = np.zeros((52, 3), np.uint8)
    if deck_csv is not None:
        deck = np.genfromtxt(deck_csv, delimiter=",")
    else:
        deck = get_shuffled_deck()
    _move_from_old_to_new(deck, game, 0, 0)
    game_index = 1

    if interactive:
        console.rule("[b blue]Welcome to interactive mode![/b blue]")
        console.print(
            f"[b blue] In this mode the program stops every time it "
            f"evaluates a card for a match.[/b blue]"
        )
        console.print(
            f"[b blue] The card/stack being evaluated for a match is surrounded by "
            f"blue bars like so:[/b blue]"
            f"[default on blue] [/default on blue]"
            f"[{SUIT_STYLES[2]}] A♥ [/{SUIT_STYLES[2]}]"
            f"[default on blue] [/default on blue]"
        )
        console.print(
            "[b blue] The number of cards in a stack including the top "
            "card is indicated by the green number, for example:[/b blue]"
            "([green]4[/green])"
        )
        console.print(
            f"[b blue] The last stack to have a card/stack placed on it is "
            f"surrounded by red bars like so:[/b blue] "
            f"[default on red] [/default on red]"
            f"[{SUIT_STYLES[2]}] A♥ [/{SUIT_STYLES[2]}]"
            f"[default on red] [/default on red]"
        )
        input("Press enter to continue\n")

    last_stacked = None
    for i in range(1, 52):
        _move_from_old_to_new(deck, game, i, game_index)
        if interactive:
            console.print(f"[u]Cards left in deck: {51 - i}[u]")
        last_stacked, game_index = combine_recursive(
            game, game_index, game_index, game_index, interactive, last_stacked
        )
        game_index += 1

    game_index -= 1

    if print_game:
        print_current_state(game, game_index)

    if save_csv:
        if isinstance(save_csv, str):
            np.savetxt(save_csv, deck, delimiter=",")
        else:
            ValueError("save_csv must be False or str type")

    return game[0][TOTAL_CARDS]


import datetime
import itertools
import time
from typing import Optional

import dask
from dask import delayed
from dask.distributed import Client

from cg_simulator.core import console
from cg_simulator.games.accordion import multiple_simulations, simulate_recursive


def run_simulation(
    print_game: bool = False,
    interactive: bool = False,
    cost_per_deck: float = 0,
    earned_per_card: float = 0,
    save_csv: Optional[str] = None,
    deck_csv: Optional[str] = None,
) -> dict:
    print_game = print_game or interactive
    total_cards = simulate_recursive(print_game, interactive, save_csv, deck_csv)

    results = {"total_cards": total_cards}
    gross = total_cards * earned_per_card
    net = gross - cost_per_deck

    if print_game:
        console.print(f"Cards in your first pile: {total_cards}")

    if cost_per_deck or earned_per_card:
        results["gross"] = gross
        results["net"] = net
        if print_game:
            net_color = "red" if net < 0 else "green"
            console.print(
                f"You earned ${gross: ,.2f} [blue]({total_cards} cards x "
                f"${earned_per_card: ,})[/blue] from the game"
            )
            console.print(f"A deck of cards cost: ${cost_per_deck: ,.2f}")
            console.print(
                f"Making your net total: [{net_color}]${net: ,.2f}[/{net_color}]"
            )

    return results


def run_multiple_simulations(
    simulations: int = 10,
    print_results: bool = False,
    cost_per_deck: float = 0,
    earned_per_card: float = 0,
    print_games: bool = False,
    use_dask: bool = False,
    dask_address: Optional[str] = None,
    dask_chunk: int = 1_000_000,
    custom_dask_client: Optional[Client] = None,
) -> dict:
    if use_dask:
        if custom_dask_client is not None:
            client = custom_dask_client
            if dask_address:
                console.print(
                    f"[red]Since a custom dask client was provided, the dask address "
                    f"of[/red] {dask_address} [red]is being ignored[/red]"
                )
        else:
            client = Client(dask_address)
        console.print(client)

        if print_games:
            console.print("[red]Printing games is disabled when using dask.[/red]")
            console.print("[blue]Final results will still be printed[/blue]")

        tasks = simulations // dask_chunk
        left_over = simulations % dask_chunk

        results = [delayed(multiple_simulations)(dask_chunk)] * tasks
        results.append(delayed(multiple_simulations)(left_over))

        results = dask.compute(*results)
        results = list(itertools.chain(*results))

    elif print_games:
        results = [simulate_recursive(True) for _ in range(simulations)]

    else:
        console.print("[yellow]Running Simulations")
        start = time.time()
        results = multiple_simulations(simulations)
        end = time.time()
        time_taken = str(datetime.timedelta(seconds=round(end - start, 2)))[:-4]
        console.print(f"[green]Finished running Simulations in {time_taken}")

    total_cards = sum(results)
    average_cards = total_cards / simulations
    print_results = print_results or print_games
    final_results = {
        "results": results,
        "total_cards": total_cards,
        "average_cards": average_cards,
    }

    if print_results:
        console.rule("[b]RESULTS[/b]")
        console.print(f"You played {simulations: ,} games")
        console.print(f"Average cards in first stack: {average_cards}")

    if cost_per_deck or earned_per_card:
        gross_total = total_cards * earned_per_card
        net_total = gross_total - cost_per_deck * simulations

        gross_average = gross_total / simulations
        net_average = net_total / simulations

        final_results["gross_total"] = gross_total
        final_results["net_total"] = net_total
        final_results["gross_average"] = gross_average
        final_results["net_average"] = net_average

        if print_results:
            net_color = "red" if net_total < 0 else "green"
            console.print(f"Total gross earnings: [green]${gross_total: ,.2f}[/green]")
            console.print(
                f"Total net earnings: [{net_color}]${net_total: ,.2f}[/{net_color}]"
            )
            console.print(
                f"Average gross earnings: [green]${gross_average: ,.2f}[/green]"
            )
            console.print(
                f"Average net earnings: "
                f"[{net_color}]${net_average: ,.2f}[/{net_color}]"
            )

    return final_results



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cg_simulator', 'cg_simulator.games']

package_data = \
{'': ['*']}

install_requires = \
['dask[distributed]>=2021.4.1,<2022.0.0',
 'numba>=0.53.1,<0.54.0',
 'numpy>=1.20.2,<2.0.0',
 'rich>=10.1.0,<11.0.0',
 'typer>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['accordion = accordion_simulator.cli:simulate']}

setup_kwargs = {
    'name': 'cg-simulator',
    'version': '0.1.0',
    'description': 'A simulator for card games.',
    'long_description': None,
    'author': 'Raymond DeVries',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)
